import os
import sys
import time
import unittest

from AnyQt.QtGui import QPalette
from AnyQt.QtTest import QSignalSpy

from orangecanvas.utils import shtools as sh
from orangecanvas.application import application as appmod
from orangecanvas.utils.shtools import temp_named_file


def application_test_helper():
    app = appmod.CanvasApplication([])
    p = app.palette()
    spy = QSignalSpy(app.applicationPaletteChanged)
    p.setColor(QPalette.Base, p.color(QPalette.Text))
    app.setPalette(p)
    assert list(spy) == [[]]
    app.quit()
    return


class TestApplication(unittest.TestCase):
    def test_application(self):
        res = sh.python_run([
            "-c",
            f"import {__name__} as m\n"
            f"m.application_test_helper()\n"
        ])
        self.assertEqual(res.returncode, 0)

    def test_application_help(self):
        res = sh.python_run([
            "-m", "orangecanvas", "--help"
        ])
        self.assertEqual(res.returncode, 0)


def remove_after_exit(fname):
    appmod.run_after_exit([
        sys.executable, '-c', f'import os, sys; os.remove(sys.argv[1])', fname
    ])


def restart_command_test_helper(fname):
    cmd = [
        sys.executable, '-c', f'import os, sys; os.remove(sys.argv[1])', fname
    ]
    appmod.set_restart_command(cmd)
    assert appmod.restart_command() == cmd
    appmod.restart_cancel()
    assert appmod.restart_command() is None
    appmod.set_restart_command(cmd)


class TestApplicationRestart(unittest.TestCase):
    def test_restart_command(self):
        with temp_named_file('', delete=False) as fname:
            res = sh.python_run([
                "-c",
                f"import sys, {__name__} as m\n"
                f"m.restart_command_test_helper(sys.argv[1])\n",
                fname
            ])
            start = time.perf_counter()
            while os.path.exists(fname) and time.perf_counter() - start < 5:
                pass
            self.assertFalse(os.path.exists(fname))
            self.assertEqual(res.returncode, 0)



def plsi(n: int, forms: str) -> str:
    """
    Choose a plural form for Slovenian - or create one, for some rare cases

    `forms` can be a string containing the singular and plural form, separated
    by "|", for instance `"okno|okni|okna|oken".

    The number of forms must be 4 or 3.
    - The four forms are singular, dual, plural for 3 or 4, plural for >= 5
    - Three forms are used for cases other than genitive, where plural is the
      same for all numbers >= 3

    A single form can be given for nouns in genitive that conform to one of
    the following rules:
    - miza/mizi/mize/miz
    - korak/koraka/koraki/korakov
    The function does not speak Slovenian and cannot verify the conformance. :)

    Examples:

    - Four plural forms:
        f'Aktiven {nfilt} {plsi(n, "filter|filtra|filtri|filtrov")}'

    - Four forms, multiple words conjugated:
        f'V tabeli je {n} {plsi(n, "učni primer|učna primera|učni primeri|učnih primerov")}'

    - Three forms (non-nominative):
        f'Datoteka z {n} {plsi(n, "primerom|primeroma|primeri")}'

    - Single form, feminine, using pattern
        f'Najdena {nvars} {plsi(nvars, "spremenljivka")}'

    - Single form, masculine, using pattern
        f'Vsebina: {n} {plsi(n, "primer")'

    - Plural form used twice
        f'{plsi(n, "Ostalo je|Ostala sta|Ostali so")} še {n} {plsi(n, "primer")}'

    Args:
        n: number
        forms: plural forms, separated by "|", or a single (regular) noun

    Returns:
        form corresponding to the given number
    """
    n = abs(n) % 100
    if n == 4:
        n = 3
    elif n == 0 or n >= 5:
        n = 4
    n -= 1

    if "|" in forms:
        forms = forms.split("|")
        # Don't use max: we want it to fail if there are just two forms
        if n == 3 and len(forms) == 3:
            n -= 1
        return forms[n]

    if forms[-1] == "a":
        return forms[:-1] + ("a", "i", "e", "")[n]
    else:
        return forms + ("", "a", "i", "ov")[n]


def plsi_sz(n: int) -> str:
    """
    Returns proposition "s" or "z", depending on the number that will follow it.

    Args:
        n (int): number

    Returns:
        Proposition s or z
    """
    # Cut of all groups of three, except the first one
    lead3 = f"{n:_}".split("_")[0]

    # handle 1, 1_XXX, 1_XXX_XXX ... because "ena" is not pronounced and we need
    # to match "tisoč", "milijon", ... "trilijarda"
    # https://sl.wikipedia.org/wiki/Imena_velikih_%C5%A1tevil
    if lead3 == "1":
        if n > 10 ** 63:  # nobody knows their names
            return "z"
        return "zszzzzsssssssssszzzzzz"[len(str(n)) // 3]

    # This is pronounced sto...something
    if len(lead3) == 3 and lead3[0] == "1":
        return "s"

    # Take the first digit, or the second for two-digit number not divisible by 10
    lead = lead3[len(lead3) == 2 and lead3[1] != "0"]
    return "zzzssssszz"[int(lead)]


def z_besedo(n, case, gender, zero="nič"):
    if not 0 <= n <= 10:
        return str(n)
    if n == 0:
        return zero
    elif n == 1:
        return {"m": ("", "en", "enega", "enemu", "en", "enem", "enim"),
                "f": ("", "ena", "ene", "eni", "eno", "eni", "eno"),
                "n": ("", "eno", "enega", "enemu", "eno", "enem", "enim")}[gender][case]
    elif n == 2:
        return {"m": ("", "dva", "dveh", "dvema", "dva", "dveh", "dvema"),
                "f": ("", "dve", "dveh", "dvema", "dve", "dveh", "dvema"),
                "n": ("", "dve", "dveh", "dvema", "dve", "dveh", "dvema")}[gender][case]
    return (None, None, None,
            ("", "tri", "treh", "trem", "tri", "treh", "tremi"),
            ("", "štiri", "štirih", "štirim", "štiri", "štirih", "štirimi"),
            ("", "pet", "petih", "petim", "pet", "petih", "petimi"),
            ("", "šest", "šestih", "šestim", "šest", "šestih", "šestimi"),
            ("", "sedem", "sedmih", "sedmim", "sedem", "sedmih", "sedmimi"),
            ("", "osem", "osmih", "osmim", "osem", "osmih", "osmimi"),
            ("", "devet", "devetih", "devetim", "devet", "devetih", "devetimi"),
            ("", "deset", "desetih", "desetim", "deset", "desetih", "desetimi"))[n][case]



from functools import lru_cache
import warnings

import os
import json
import importlib

from AnyQt.QtCore import QSettings, QLocale

def pl(n: int, forms: str) -> str:  # pylint: disable=invalid-name
    """
    Choose a singular/plural form for English - or create one, for regular nouns

    `forms` can be a string containing the singular and plural form, separated
    by "|", for instance `pl(n, "leaf|leaves")`.

    For nouns that are formed by adding an -s (e.g. tree -> trees),
    and for nouns that end with -y that is replaced by -ies
    (dictionary -> dictionaries), it suffices to pass the noun,
    e.g. `pl(n, "tree")`, `pl(n, "dictionary")`.

    Args:
        n: number
        forms: plural forms, separated by "|", or a single (regular) noun

    Returns:
        form corresponding to the given number
    """
    plural = int(n != 1)

    if "|" in forms:
        return forms.split("|")[plural]

    if forms[-1] in "yY" and forms[-2] not in "aeiouAEIOU":
        word = [forms, forms[:-1] + "ies"][plural]
    else:
        word = forms + "s" * plural
    if forms.isupper():
        word = word.upper()
    return word

def _load_json(path):
    with open(path) as handle:
        return json.load(handle)

@lru_cache
def get_languages(package=None):
    if package is None:
        package = "orangecanvas"
    package_path = os.path.dirname(importlib.import_module(package).__file__)
    msgs_path = os.path.join(package_path, "i18n")
    if not os.path.exists(msgs_path):
        return {}
    names = {}
    for name, ext in map(os.path.splitext, os.listdir(msgs_path)):
        if ext == ".json":
            try:
                msgs = _load_json(os.path.join(msgs_path, name + ext))
            except json.JSONDecodeError:
                warnings.warn("Invalid language file "
                              + os.path.join(msgs_path, name + ext))
            else:
                names[msgs[0]] = name
    return names


DEFAULT_LANGUAGE = QLocale().languageToString(QLocale().language())
if DEFAULT_LANGUAGE not in get_languages():
    DEFAULT_LANGUAGE = "English"


def language_changed():
    s = QSettings()
    lang = s.value("application/language", DEFAULT_LANGUAGE)
    last_lang = s.value("application/last-used-language", DEFAULT_LANGUAGE)
    return lang != last_lang


def update_last_used_language():
    s = QSettings()
    lang = s.value("application/language", "English")
    s.setValue("application/last-used-language", lang)


class _list(list):
    # Accept extra argument to allow for the original string
    def __getitem__(self, item):
        if isinstance(item, tuple):
            item = item[0]
        return super().__getitem__(item)


class Translator:
    e = eval

    def __init__(self, package, organization="biolab.si", application="Orange"):
        s = QSettings(QSettings.IniFormat, QSettings.UserScope,
                      organization, application)
        lang = s.value("application/language", DEFAULT_LANGUAGE)
        # For testing purposes (and potential fallback)
        # lang = os.environ.get("ORANGE_LANG", "English")
        package_path = os.path.dirname(importlib.import_module(package).__file__)
        lang_eng = get_languages().get(lang, lang)
        path = os.path.join(package_path, "i18n", f"{lang_eng}.json")
        if not os.path.exists(path):
            path = os.path.join(package_path, "i18n", f"{DEFAULT_LANGUAGE}.json")
        assert os.path.exists(path), f"Missing language file {path}"
        self.m = _list(_load_json(path))

    # Extra argument(s) can give the original string or any other relevant data
    def c(self, idx, *_):
        return compile(self.m[idx], '<string>', 'eval')



"""
Test for splashscreen
"""
import pkgutil
from datetime import datetime

from AnyQt.QtGui import QPixmap, QImage
from AnyQt.QtCore import Qt, QRect, QTimer

from ..splashscreen import SplashScreen

from ..test import QAppTestCase
from ... import config


class TestSplashScreen(QAppTestCase):
    def test_splashscreen(self):
        contents = pkgutil.get_data(
            config.__package__, "icons/orange-canvas-core-splash.svg"
        )
        img = QImage.fromData(contents)
        w = SplashScreen()
        w.setPixmap(QPixmap.fromImage(img))
        w.setTextRect(QRect(100, 100, 400, 50))
        w.show()

        def advance_time():
            now = datetime.now()
            time = now.strftime("%c : %f")
            i = now.second % 3
            if i == 2:
                w.setTextFormat(Qt.RichText)
                time = "<i>" + time + "</i>"
            else:
                w.setTextFormat(Qt.PlainText)

            w.showMessage(time, alignment=Qt.AlignCenter)

            rect = QRect(100, 100 + i * 20, 400, 50)
            w.setTextRect(rect)

            self.assertEqual(w.textRect(), rect)

        timer = QTimer(w, interval=1)
        timer.timeout.connect(advance_time)
        timer.start()
        self.qWait()
        timer.stop()


import os
import sys
import unittest

# This test will usually not be run when the module `si` is embedded within
# Orange structure, hence we manually insert its path
from itertools import chain

si_path = os.path.join(os.path.dirname(__file__), "..")
sys.path.append(si_path)
from si import plsi, plsi_sz, z_besedo


class TestPlsi(unittest.TestCase):
    def test_plsi_4(self):
        self.assertEqual(plsi(0, "okno|okni|okna|oken"), "oken")
        self.assertEqual(plsi(1, "okno|okni|okna|oken"), "okno")
        self.assertEqual(plsi(2, "okno|okni|okna|oken"), "okni")
        self.assertEqual(plsi(3, "okno|okni|okna|oken"), "okna")
        self.assertEqual(plsi(4, "okno|okni|okna|oken"), "okna")
        self.assertEqual(plsi(5, "okno|okni|okna|oken"), "oken")
        self.assertEqual(plsi(11, "okno|okni|okna|oken"), "oken")
        self.assertEqual(plsi(100, "okno|okni|okna|oken"), "oken")
        self.assertEqual(plsi(101, "okno|okni|okna|oken"), "okno")
        self.assertEqual(plsi(102, "okno|okni|okna|oken"), "okni")
        self.assertEqual(plsi(103, "okno|okni|okna|oken"), "okna")
        self.assertEqual(plsi(105, "okno|okni|okna|oken"), "oken")
        self.assertEqual(plsi(1001, "okno|okni|okna|oken"), "okno")

    def test_plsi_3(self):
        self.assertEqual(plsi(0, "oknu|oknoma|oknom"), "oknom")
        self.assertEqual(plsi(1, "oknu|oknoma|oknom"), "oknu")
        self.assertEqual(plsi(2, "oknu|oknoma|oknom"), "oknoma")
        self.assertEqual(plsi(3, "oknu|oknoma|oknom"), "oknom")
        self.assertEqual(plsi(5, "oknu|oknoma|oknom"), "oknom")
        self.assertEqual(plsi(1, "oknu|oknoma|oknom"), "oknu")
        self.assertEqual(plsi(105, "oknu|oknoma|oknom"), "oknom")

    def test_plsi_1(self):
        self.assertEqual(plsi(0, "miza"), "miz")
        self.assertEqual(plsi(1, "miza"), "miza")
        self.assertEqual(plsi(2, "miza"), "mizi")
        self.assertEqual(plsi(3, "miza"), "mize")
        self.assertEqual(plsi(5, "miza"), "miz")
        self.assertEqual(plsi(101, "miza"), "miza")
        self.assertEqual(plsi(105, "miza"), "miz")

        self.assertEqual(plsi(0, "primer"), "primerov")
        self.assertEqual(plsi(1, "primer"), "primer")
        self.assertEqual(plsi(2, "primer"), "primera")
        self.assertEqual(plsi(3, "primer"), "primeri")
        self.assertEqual(plsi(5, "primer"), "primerov")
        self.assertEqual(plsi(50, "primer"), "primerov")
        self.assertEqual(plsi(101, "primer"), "primer")
        self.assertEqual(plsi(105, "primer"), "primerov")


class TestPlsi_si(unittest.TestCase):
    def test_plsi_sz(self):
        for propn in "z0 z1 z2 s3 s4 s5 s6 s7 z8 z9 z10 " \
                      "z11 z12 s13 s14 s15 s16 s17 z18 z19 z20 " \
                      "z21 z22 s23 z31 z32 s35 s40 s50 s60 s70 z80 z90 " \
                      "z200 z22334 s3943 z832492 " \
                      "s100 s108 s1000 s13333 s122222 z1000000 " \
                      "z1000000000 z1000000000000".split():
            self.assertEqual(plsi_sz(int(propn[1:])), propn[0], propn)


class TestZBesedo(unittest.TestCase):
    def test_z_besedo(self):
        self.assertEqual(
            "\n".join(
                f"{z_besedo(n, 1, 'n')} "
                f"{plsi(n, 'zeleno drevo|zeleni drevesi|zelena drevesa|zelenih dreves')}"
                for n in chain(range(6), (11, ))),
            """nič zelenih dreves
eno zeleno drevo
dve zeleni drevesi
tri zelena drevesa
štiri zelena drevesa
pet zelenih dreves
11 zelenih dreves"""
        )

        self.assertEqual(
            "\n".join(
                f"{plsi_sz(n).upper()} {z_besedo(n, 6, 'n')} "
                f"{plsi(n, 'zelenim drevesom|zelenima drevesoma|zelenimi drevesi')}"
                for n in chain(range(1, 6), (11, 100, 101))),
            """Z enim zelenim drevesom
Z dvema zelenima drevesoma
S tremi zelenimi drevesi
S štirimi zelenimi drevesi
S petimi zelenimi drevesi
Z 11 zelenimi drevesi
S 100 zelenimi drevesi
S 101 zelenim drevesom""")

        self.assertEqual(f"{z_besedo(0, 1, 'm', 'brez')} {plsi(0, 'drevesa|dreves|dreves')}",
                         "brez dreves")


sys.path.remove(si_path)

if __name__ == "__main__":
    unittest.main()



import os
import sys
import re
import email
from operator import itemgetter
from typing import List, Dict, Optional, Union, cast

import packaging.version

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint, Distribution, entry_points, PackageNotFoundError
else:
    from importlib.metadata import EntryPoint, Distribution, entry_points, PackageNotFoundError


__all__ = [
    "Distribution", "EntryPoint", "entry_points", "normalize_name", "trim",
    "trim_leading_lines", "trim_trailing_lines", "parse_meta", "get_dist_meta",
    "get_distribution"
]


def normalize_name(name: str) -> str:
    """
    PEP 503 normalization plus dashes as underscores.
    """
    return re.sub(r"[-_.]+", "-", name).lower().replace('-', '_')


def is_develop_egg(dist: Distribution) -> bool:
    """
    Is the distribution installed in development mode (setup.py develop)
    """
    egg_info_dir = dist.locate_file(f"{normalize_name(dist.name)}.egg-info/")
    setup = dist.locate_file("setup.py")
    return os.path.isdir(egg_info_dir) and os.path.isfile(setup)


def left_trim_lines(lines: List[str]) -> List[str]:
    """
    Remove all unnecessary leading space from lines.
    """
    lines_striped = zip(lines[1:], map(str.lstrip, lines[1:]))
    lines_striped = filter(itemgetter(1), lines_striped)
    indent = min([len(line) - len(striped)
                  for line, striped in lines_striped] + [sys.maxsize])

    if indent < sys.maxsize:
        return [line[indent:] for line in lines]
    else:
        return list(lines)


def trim_trailing_lines(lines: List[str]) -> List[str]:
    """
    Trim trailing blank lines.
    """
    lines = list(lines)
    while lines and not lines[-1]:
        lines.pop(-1)
    return lines


def trim_leading_lines(lines: List[str]) -> List[str]:
    """
    Trim leading blank lines.
    """
    lines = list(lines)
    while lines and not lines[0]:
        lines.pop(0)
    return lines


def trim(string: str) -> str:
    """
    Trim a string in PEP-256 compatible way
    """
    lines = string.expandtabs().splitlines()
    lines = list(map(str.lstrip, lines[:1])) + left_trim_lines(lines[1:])
    return "\n".join(trim_leading_lines(trim_trailing_lines(lines)))


# Fields allowing multiple use (from PEP-0345)
MULTIPLE_KEYS = ["Platform", "Supported-Platform", "Classifier",
                 "Requires-Dist", "Provides-Dist", "Obsoletes-Dist",
                 "Project-URL"]


def parse_meta(contents: str) -> Dict[str, Union[str, List[str]]]:
    message = email.message_from_string(contents)
    meta = {}  # type: Dict[str, Union[str, List[str]]]
    for key in set(message.keys()):
        if key in MULTIPLE_KEYS:
            meta[key] = list(str(m) for m in message.get_all(key, []))
        else:
            value = str(message.get(key))
            if key == "Description":
                value = trim(value)
            meta[key] = value
    version_str = cast(str, meta["Metadata-Version"])
    version = packaging.version.parse(version_str)
    if version >= packaging.version.parse("1.3") and "Description" not in meta:
        desc = message.get_payload()
        if isinstance(desc, str):
            meta["Description"] = desc
    return meta


def get_meta_entry(dist: Distribution, name: str) -> Union[List[str], str, None]:
    """
    Get the contents of the named entry from the distributions PKG-INFO file
    """
    meta = get_dist_meta(dist)
    return meta.get(name)


def get_dist_url(dist: Distribution) -> Optional[str]:
    """
    Return the 'url' of the distribution (as passed to setup function)
    """
    url = get_meta_entry(dist, "Home-page")
    assert isinstance(url, str) or url is None
    return url


def get_dist_meta(dist: Distribution) -> Dict[str, Union[str, List[str]]]:
    metadata = dist.metadata
    meta: Dict[str, Union[str, List[str]]] = {}
    for key in metadata:
        if key == "Description":
            meta[key] = trim(metadata[key])
        elif key in MULTIPLE_KEYS:
            meta[key] = metadata.get_all(key)
        else:
            meta[key] = metadata[key]
    return meta


def get_distribution(name: str) -> Optional[Distribution]:
    try:
        return Distribution.from_name(name)
    except PackageNotFoundError:
        return None
